! main.f90
!
! Program to compute the transitions for the paper
!    "The Macroeconomic Impact of Adding Liquidity Regulations to
!            Bank Capital Regulations"
!
! Francisco B. Covas
! Washington DC, Winter 2012/2013
!

program transition

   use mod_types
   use mod_common_var
   use model_subs
   
   implicit none 

   include 'mpif.h'

   integer :: i,j,n,jz,m,iz,jx,jj
   integer :: nvar,iter,ios
   integer :: SolveFlag,ErrorFlag
   real (dp) :: cover_tauchen,kss,x,dapc,aux,fnorm
   real (dp) :: wvars(7),evars(9),bvars(21),supn(3)
   real (dp) :: Lc,Kc,Yc,Yagg,Cagg
   real (dp), allocatable, dimension(:) :: xv,xv1 
   real (dp), allocatable, dimension(:,:) :: xfjac,invxfjac,d,fval,fold,u,dpr

   !--------------------------------------------------------------------------------
   ! 1.  Initialize MPI, get rank and # of processes
   !---------------+-----------------------------------------------------------------
   
   call MPI_Init(ierr)
   call MPI_Comm_rank(MPI_COMM_WORLD, id, ierr)
   call MPI_Comm_size(MPI_COMM_WORLD, nproc, ierr)
        
   print *, 'hello, world! i am node ',id

   call cpu_time(time_begin)
      
   !---------------------------------------------------------------------------
   ! 2.  Tauchen points
   !---------------------------------------------------------------------------

   cover_tauchen = 3.d0

   ! Workers

   allocate(wbar(nzw),wtm(nzw,nzw))
   call sub_tauchen (wbar,wtm,rho_w,sigma_w,nzw,cover_tauchen)
   wbar = exp(wbar)

   if (id == idmom) then
      print '(" ")'
      print '(" Workers: states of the Markov Chain ")' 
      print '(" ")'
      print '("     Earnings ability ")'
      print '(" ")'
      print '(30f10.6)',wbar
      print '(" ")'
      print '("    Transition Matrix       ")' 
      print '(" ")'
      do jz=1,nzw
         print '(30f10.6)',wtm(jz,:)
      end do
      print '(" ")'
   end if

   ! Entrepreneurs

   allocate(ebar(nze),etm(nze,nze))
   call sub_tauchen (ebar,etm,rho_e,sigma_e,nze,cover_tauchen)
   ebar = exp(ebar)
   allocate(zesim(nsims))
   call simulate_markov(nze,nsims,etm,z0,zesim)

   if (id == idmom) then
      print '(" ")'
      print '(" Entrepreneurs: states of the Markov Chain ")' 
      print '(" ")'
      print '("     Productivity ")'
      print '(" ")'
      print '(30f10.6)',ebar
      print '(" ")'
      print '("    Transition Matrix       ")' 
      print '(" ")'
      do jz=1,nze
         print '(30f10.6)',etm(jz,:)
      end do
      print '(" ")'
   end if

   delte1 = (1.0_dp-delte)
   alphe1 = (alphe-1.0_dp)
   alpha1 = (alpha-1.0_dp)
   ne1    = 1.0_dp/(1.0_dp-ne)
   alph1  = (1-alpha)
   delta1 = (1-delta)
   nu_b   = (1.0_dp-nu_e-nu_w)
   rk     = 0.065_dp
   rb     = 0.0013_dp
   rd     = 0.0013_dp
   rg     = 0.0080_dp
   grg    = (1.0_dp+rg)
   grb    = (1.0_dp+rb)
   mgr    = (1.0_dp+rk-phi)
 
   ! Bankers

   ! Risky technology
   allocate(y(nz),tm(nz,nz))
   call sub_tauchen (y,tm,rho,sigma,nz,cover_tauchen)
   y = dexp(y)
   allocate(zsim(nsims))
   call simulate_markov(nz,nsims,tm,z0,zsim)

   ! Deposits technology
   allocate(dst(nst),st(nst),tm_st(nst,nst))
   call sub_tauchen (dst,tm_st,rho_st,sigma_st,nst,cover_tauchen)
   allocate(stsim(nsims))
   call simulate_markov(nst,nsims,tm_st,st0,stsim)

   if (id == idmom) then
      print '(" ")'
      print '(" States of the Markov Chain ")' 
      print '(" ")'
      print '("     Risky Loans ")'
      print '(" ")'
      print '(7f10.6)',y
      print '(" ")'
      print '("    Transition Matrix       ")' 
      print '(" ")'
      do jz=1,nz
         print '(7f10.6)',tm(jz,:)
      end do
      print '(" ")'
      print '("          Deposits  ")'
      print '(" ")'
      print '(7f10.6)',dexp(dst)
      print '(" ")'
      print '("    Transition Matrix       ")' 
      print '(" ")'
      do jz=1,nst
         print '(7f10.6)',tm_st(jz,:)
      end do
      print '(" ")'
   end if 

   !---------------------------------------------------------------------------
   ! 3.  Continuous state variables 
   !---------------------------------------------------------------------------

   ! Workers

   nsmprdw(1) = 1            ! partial products of nsd
   nsmsumw(1) = 0            ! partial sums of nsd
   nsdsumw    = nsdw(1)      ! sums nsd
   nspw = nsmprdw(1)*nsdw(1) ! number of grid points

   allocate(smw(nsdsumw))
   n=1
   smw(n) = wmin
   do i=2,nspw
      n = n+1
      smw(n) = wmin+(wmax-wmin)*(dfloat(i-1)/dfloat(nsdw(1)-1))**suw
   end do

   if (id == idmom) then 
      !print '(" ")'
      !print '(" Workers: Grid for wealth ")'
      !print '(" ")'
      !print '(10f6.2)',( smw(m), m=1,nspw )
      !print '(" ")'
   end if

   ! Entrepreneurs

   nsmprde(1) = 1            ! partial products of nsd
   nsmsume(1) = 0            ! partial sums of nsd
   nsdsume    = nsde(1)      ! sums nsd
   nspe = nsmprde(1)*nsde(1) ! number of grid points

   allocate(sme(nsdsume))
   n=1
   sme(n) = emin
   do i=2,nspe
      n = n+1
      sme(n) = emin+(emax-emin)*(dfloat(i-1)/dfloat(nsde(1)-1))**sue
   end do

   if (id == idmom) then 
      !print '(" ")'
      !print '(" Entrepreneurs: Grid for wealth ")'
      !print '(" ")'
      !print '(10f8.2)',( sme(m), m=1,nspe )
      !print '(" ")'
   end if
   
   ! Bankers

   !--> Setting up the state space:
   nsmprd(1) = 1       ! partial products of nsd
   nsmsum(1) = 0       ! partial sums of nsd
   nsdsum    = nsd(1)  ! sums nsd
   do i=2,ns
      nsmprd(i) = nsmprd(i-1)*nsd(i-1)
      nsmsum(i) = nsmsum(i-1)+nsd(i-1)
      nsdsum    = nsdsum+nsd(i)
   end do
   nsp = nsmprd(ns)*nsd(ns)
 
   !--> Grid points for net assets and loans
   allocate(sm(nsdsum))
   n=0
   do j=1,ns
      n = n+1
      sm(n) = smin(j)
      do i=2,nsd(j)
         n = n+1
         sm(n) = smin(j)+(smax(j)-smin(j))*(dfloat(i-1)/dfloat(nsd(j)-1))**suv(j)
      end do
   end do

   allocate(smins(ns,nz))
   allocate(smaxs(ns,nz))
   smins(1,:) = smin(1) !(/   1.4,   1.4,   1.4,   1.4,   1.4 /)
   smins(2,:) = smin(2) !(/   3.0,   3.0,   3.0,   3.0,   3.0 /)
   smaxs(1,:) = smax(1) !(/ 120.0, 120.0, 120.0, 120.0, 120.0 /)
   smaxs(2,:) = smax(2) !(/ 260.0, 260.0, 260.0, 260.0, 260.0 /)

   allocate(sms(nsdsum,nz))
   do jj=1,nz
      n=0
      do j=1,ns
         n = n+1
         sms(n,jj) = smins(j,jj)
         do i=2,nsd(j)
            n = n+1
            sms(n,jj) = smins(j,jj)+(smaxs(j,jj)-smins(j,jj))*(dfloat(i-1)/dfloat(nsd(j)-1))**suv(j)
         end do
      end do
   end do 

   if (id == idmom) then 
      !print '(" ")'
      !print '(" Grid for net assets ")'
      !print '(" ")'
      !print '(10f6.2)',( sm(m), m=1,nsd(1) )
      !print '(" ")'
      !print '(" Grid for loans ")'
      !print '(" ")'
      !print '(10f6.2)',( sm(m), m=nsd(1)+1,nsdsum )
      !print '(" ")'
   end if

   !---------------------------------------------------------------------------
   ! 4.  Initial guess 
   !---------------------------------------------------------------------------
  
   ! Workers

   allocate(hwpm(nspw,nzw))
   allocate(hwm(nspw,nzw))
   allocate(hdwt(nspw,nzw))
   allocate(hawt(nspw,nzw))
   allocate(wctp(nspw,nzw,nper))
   allocate(wdtp(nspw,nzw,nper))
   allocate(watp(nspw,nzw,nper))
   
   open(unit=11,file='wcons.dat')
   do iz=1,nzw 
      read(11,'(100000f20.10)') ( hwm(m,iz), m=1,nspw )
   end do
   close(11)
   open(unit=11,file='wdeposits.dat')
   do iz=1,nzw 
      read(11,'(100000f20.10)') ( hdwt(m,iz), m=1,nspw )
   end do
   close(11)
   open(unit=11,file='wassets.dat')
   do iz=1,nzw 
      read(11,'(100000f20.10)') ( hawt(m,iz), m=1,nspw )
   end do
   close(11)

   do j=1,nper
      wctp(:,:,j) = hwm
      wdtp(:,:,j) = hdwt
      watp(:,:,j) = hawt
   end do

   ! Entrepreneurs

   allocate(hepm(nspe,nze))
   allocate(hem(nspe,nze))
   allocate(hket(nspe,nze))
   allocate(hset(nspe,nze))
   allocate(hbet(nspe,nze))
   allocate(ectp(nspe,nze,nper))
   allocate(ektp(nspe,nze,nper))
   allocate(ebtp(nspe,nze,nper))
   allocate(eatp(nspe,nze,nper))

   open(unit=11,file='econs.dat')
   do iz=1,nze 
      read(11,'(100000f20.10)') ( hem(m,iz), m=1,nspe )
   end do
   close(11)
   open(unit=11,file='ecapital.dat')
   do iz=1,nze 
      read(11,'(100000f20.10)') ( hket(m,iz), m=1,nspe )
   end do
   close(11)
   open(unit=11,file='edebt.dat')
   do iz=1,nze 
      read(11,'(100000f20.10)') ( hbet(m,iz), m=1,nspe )
   end do
   close(11)
   open(unit=11,file='eassets.dat')
   do iz=1,nze 
      read(11,'(100000f20.10)') ( hset(m,iz), m=1,nspe )
   end do
   close(11)

   do j=1,nper  
      ectp(:,:,j) = hem
      ektp(:,:,j) = hket
      ebtp(:,:,j) = hbet
      eatp(:,:,j) = hset
   end do

   ! Bankers

   allocate(vf(nsp,nz,nst))
   allocate(value(nsp,nz,nst))
   allocate(g_k(nsp,nz,nst))
   allocate(g_s(nsp,nz,nst))
   allocate(g_c(nsp,nz,nst))
   allocate(g_l(nsp,nz,nst))
   allocate(g_m(nsp,nz,nst))
   allocate(bctp(nsp,nz,nst,nper))
   allocate(bltp(nsp,nz,nst,nper))
   allocate(bdtp(nsp,nz,nst,nper))
   allocate(bstp(nsp,nz,nst,nper))

   open(unit=11,file='loans.dat')
   do jz=1,nst
      do iz=1,nz 
         read(11,'(100000f20.10)') ( g_k(m,iz,jz), m=1,nsp )
      end do
   end do
   close(11)

   open(unit=11,file='securities.dat')
   do jz=1,nst
      do iz=1,nz 
         read(11,'(100000f20.10)') ( g_s(m,iz,jz), m=1,nsp )
      end do
   end do
   close(11)

   open(unit=11,file='cons.dat')
   do jz=1,nst
      do iz=1,nz 
         read(11,'(100000f20.10)') ( g_c(m,iz,jz), m=1,nsp )
      end do
   end do
   close(11)

   open(unit=11,file='lambda.dat')
   do jz=1,nst
      do iz=1,nz 
         read(11,'(100000f20.10)') ( g_l(m,iz,jz), m=1,nsp )
      end do
   end do
   close(11)

   do j=1,nper
      bctp(:,:,:,j) = g_c
      bltp(:,:,:,j) = g_k
      bstp(:,:,:,j) = g_s
      bdtp(:,:,:,j) = g_l
   end do

   !---------------------------------------------------------------------------
   ! 5.  Initial guesses for prices during the transitional dynamics
   !---------------------------------------------------------------------------

   Sbar = 7.0_dp

   ! Interbank rate
   rs0 = 1.0031346386_dp
   rs1 = rs0 
   allocate(rsvec(nper),rsvec0(nper))
   rsvec(1) = rs0
   rsvec(nper) = rs1
   do i=2,nper
      rsvec(i) = rs0+(dfloat(i-1)/dfloat(nper-1))*(rs1-rs0)
   end do

   ! Loan rate
   rl0 = 1.0407360559_dp
   rl1 = rl0
   allocate(rlvec(nper),rlvec0(nper))
   rlvec(1) = rl0
   rlvec(nper) = rl1
   do i=2,nper
      rlvec(i) = rl0+(dfloat(i-1)/dfloat(nper-1))*(rl1-rl0)
   end do
  
   kl0 = 5.84318032_dp
   kl1 = kl0
   allocate(klvec(nper),klvec0(nper))
   klvec(1) = kl0
   klvec(nper) = kl1
   do i=2,nper
      klvec(i) = kl0+(dfloat(i-1)/dfloat(nper-1))*(kl1-kl0)
   end do

   ! If better guesses are available uncomment the line below
   open(unit=3,file='prices.dat',iostat=ios)
   do n=1,nper
      read(3,'(3f20.10)') rlvec(n),rsvec(n),klvec(n)
   end do  
   close(unit=3)

   if (id == idmom) then
      do i=1,nper
         wt = (1.0_dp-alphaC)*klvec(i)**alphaC
         rt = alphaC*klvec(i)**(alphaC-1.0_dp)-deltaC
      end do
   end if
 
   ! Aggregate variables
   allocate(bloan(nper))
   allocate(bsecur(nper))
   allocate(sdepo(nper))
   allocate(wasset(nper))
   allocate(dloan(nper))
   allocate(easset(nper))
   allocate(dlabor(nper))
   allocate(blabor(nper))
   allocate(loan_mkt(nper))
   allocate(whole_mkt(nper))
   allocate(klnew(nper))
   allocate(reg_req(nper,3))

   !---------------------------------------------------------------------------
   ! 6.  Wealth shocks
   !---------------------------------------------------------------------------

   shockw = 1.0_dp
   shocke = 1.0_dp
   shockb = 1.0_dp

   shockw(1)  = 1.0090_dp
   shockw(2)  = 1.0059_dp
   shocke(1)  = 1.0024_dp
   shocke(2)  = 1.0015_dp
   shockb(1)  = 0.6923_dp
   shockb(2)  = 0.7084_dp

   !---------------------------------------------------------------------------
   ! 7.  Transition Dynamics
   !---------------------------------------------------------------------------

   ! Workers
   ! Distribute the job across cpu nodes
   allocate(jobw_mpi(nzw*nspw,2))
   n = 1
   do iz=1,nzw  
      do jx=1,nspw
         jobw_mpi(n,1) = jx 
         jobw_mpi(n,2) = iz
         n = n+1  
      end do
   end do

   ! Entrepreneurs
   ! Distribute the job across cpu nodes
   allocate(jobe_mpi(nze*nspe,2))
   n = 1
   do iz=1,nze  
      do jx=1,nspe
         jobe_mpi(n,1) = jx 
         jobe_mpi(n,2) = iz
         n = n+1  
      end do
   end do

   ! Bankers
   ! Distribute the job across cpu nodes
   allocate(jobs_mpi(nz*nst,2))
   n = 1
   do it=1,nst  
      do iz=1,nz
         jobs_mpi(n,1) = iz 
         jobs_mpi(n,2) = it
         n = n+1  
      end do
   end do

   ! transition dynamics starts here
   do jj=1,50

      if (id == idmom) then
         print '("-----------------------------------------------------------------------")'
         print '("                         Iteration ",i6)',jj
         print '("-----------------------------------------------------------------------")'
      end if

      call transitional_dynamics ()

      do i=1,nper    
         loan_mkt(i)  = (-nu_e*dloan(i)-nu_b*bloan(i))
         whole_mkt(i) = (-nu_b*bsecur(i)-nu_e*easset(i)+(nu_e+nu_b)*Sbar)
         klnew(i)     = nu_w*wasset(i)/((nu_w+nu_e)-nu_e*dlabor(i))
      end do

      if (id == idmom) then
         do i=1,nper
            print '(i6,6f9.5)',i,loan_mkt(i),whole_mkt(i),klnew(i),rlvec(i)-1,rsvec(i)-1
         end do
         print '(" ")'
      end if

      ! Update prices
      rsvec0 = rsvec
      rlvec0 = rlvec
      klvec0 = klvec
      do i=2,nper-1
         rlvec(i)    = rlvec0(i)+0.003_dp*loan_mkt(i)
         rsvec(i)    = rsvec0(i)+0.003_dp*whole_mkt(i)
         klvec(i)    = 0.90_dp*klvec0(i)+0.10_dp*klnew(i)
      end do
      supn(1) = supnorm(nper,rlvec,rlvec0)
      supn(2) = supnorm(nper,rsvec,rsvec0)
      supn(3) = supnorm(nper,klvec,klvec0)

      if (id == idmom) then
         print '("   1   2   3   4   5   6   7   8   9   10   11   12   13   14   15   16   17   18   19   20")'
         print '(20f8.2)',(10000.0_dp*(rlvec(i)-rlvec0(i)),i=1,nper)
         print '(" ")'
         print '(20f8.2)',(10000.0_dp*(rsvec(i)-rsvec0(i)),i=1,nper)
         print '(" ")'
         print '(20f8.2)',(100.0_dp*(klvec(i)-klvec0(i)),i=1,nper)
         print '(" ")'
         print '(3f20.12)',supn
         print '(" ")'
      end if        

   end do 

   allocate(banker_vars(nper,13))
   allocate(entre_vars(nper,9))
   allocate(worker_vars(nper,6))
 
   call invariant_banker_final(banker_vars)
   call invariant_worker_final(worker_vars)
   call invariant_entrepreneur_final(entre_vars)

   !---------------------------------------------------------------------------
   ! 7.  Save results for plotting 
   !---------------------------------------------------------------------------

   if (id == idmom) then

      open(unit=10,file='GE_cond.dat',status='replace')
      do n=1,nper
         write(10,'(7f20.10)') nu_b*bloan(n),-nu_b*bsecur(n),-nu_e*dloan(n),&
               nu_e*easset(n),nu_e*dlabor(n),nu_w*sdepo(n),nu_w*wasset(n)
      end do  
      close(10)
 
      open(unit=10,file='prices.dat',status='replace')
      do n=1,nper
         write(10,'(3f20.10)') rlvec(n),rsvec(n),klvec(n)
      end do
      close(10)

      open(unit=10,file='worker.dat',status='replace')
      do n=1,nper
         write(10,'(100000f20.10)') worker_vars(n,:)
      end do
      close(10)

      open(unit=10,file='entrepreneur.dat',status='replace')
      do n=1,nper
         write(10,'(100000f20.10)') entre_vars(n,:)
      end do
      close(10)

      open(unit=10,file='banker.dat',status='replace')
      do n=1,nper
         write(10,'(100000f20.10)') banker_vars(n,:)
      end do
      close(10)

   end if 
     
   call cpu_time(time_end)
   if (id == idmom) write(*,130) (time_end-time_begin)
   130 format (/,' Solution completed in',f12.2,' seconds ')

   !--------------------------------------------------------------------------------
   ! 7.  End MPI
   !--------------------------------------------------------------------------------
   
   call MPI_Finalize(ierr)     

end program transition

