module mod_common_var

   use mod_types

   implicit none
   
   !-----------------------------------------------------!
   ! Calibration of the model                            !
   !-----------------------------------------------------!
   
   ! Workers problem

   real (dp), parameter :: beta_w  = 0.96
   real (dp), parameter :: gamma_w = 1.0
   real (dp), parameter :: abar    = 0.0
   real (dp), parameter :: rho_w   = 0.77
   real (dp), parameter :: sigma_w = 0.22
   real (dp), parameter :: phi_w   = 0.9170
   real (dp), parameter :: nu_w    = 0.6635

   ! Entrepreneurs problem

   real (dp), parameter :: beta_e  = 0.947
   real (dp), parameter :: alphe   = 0.46
   real (dp), parameter :: ne      = 0.35
   real (dp), parameter :: delte   = 0.10
   real (dp), parameter :: gamma_e = 1.0
   real (dp), parameter :: rho_e   = 0.70
   real (dp), parameter :: sigma_e = 0.25
   real (dp), parameter :: kappe   = 0.70
   real (dp), parameter :: nu_e    = 0.3342

   ! Corporate Sector

   real (dp), parameter :: alphaC = 0.33
   real (dp), parameter :: deltaC = 0.10

   ! Bankers problem

   real (dp), parameter :: alpha  = 0.71                 ! Loan share
   real (dp), parameter :: delta  = 0.24                 ! Maturity parameter
   real (dp), parameter :: beta   = 0.947                ! Discount factor
   real (dp), parameter :: gamma  = 1.0                  ! CRRA coefficient
   real (dp), parameter :: rho    = 0.85                 ! First-order autoregressive coefficient
   real (dp), parameter :: sigma  = 0.10                 ! Conditional standard deviation
   real (dp), parameter :: kappa  = 0.90                 ! (1-capital requirements)
   real (dp), parameter :: phi    = 0.13                 ! cost of making loans
   real (dp), parameter :: tplus  = 0.02                 ! cost of originating loans
   real (dp), parameter :: tminus = 0.04                 ! cost of reducing loans
   real (dp), parameter :: rho_st = 0.80                 ! persistence of deposits
   real (dp), parameter :: sigma_st = 0.08               ! standard deviation of deposits
   real (dp), parameter :: xi1 = 0.15                    ! Liquid loans

   !-----------------------------------------------------!
   ! Numerical parameters                                !
   !-----------------------------------------------------!

   ! Workers problem

   integer, parameter :: nzw = 14                        ! # of states of the markov chain
   integer, parameter :: nsw = 1                         ! # of continuous state variables for worker's problem
   integer, parameter :: nsdw(1) = 900                   ! # of grid points
   integer, parameter :: ncw = 2                         ! # of choice variables for the worker's problem
   real (dp), parameter :: wmin = 0.70                   ! lower bound for worker's grid
   real (dp), parameter :: wmax = 100                    ! upper bound for worker's grid
   real (dp), parameter :: suw = 2.0                     ! suw > 1 gives closer spacing near wmin
  
   ! Entrepreneurs problem

   integer, parameter :: nze = 14                        ! # of states of the markov chain
   integer, parameter :: nse = 1                         ! # of state variables for entrepreneur's problem
   integer, parameter :: nsde(1) = 900                   ! # of grid points
   integer, parameter :: nce = 2                         ! # of choice variables for the entrepreneur's problem
   real (dp), parameter :: emin = 1.5                    ! lower bound on entrepreneur's grid
   real (dp), parameter :: emax = 500                    ! upper bound on entrepreneur's grid
   real (dp), parameter :: sue = 2.0                     ! sue > 1 gives closer spacing near emin
   
   ! Bankers problem

   integer, parameter :: nz =  5                         ! # of idiosyncratic productivity states (risky loans)
   integer, parameter :: nst = 4                         ! # of idiosyncratic productivity states (deposits)
   integer, parameter :: nx = 2                          ! # of continuous endogenous state vars
   integer, parameter :: ns = nx                         ! Number of continuous state variables
   integer, parameter :: nsd(2) = (/ 100, 100 /)         ! Number of gridpoints for continuous state variables
   integer, parameter :: maxit  = 1000                   ! Maximum number of iterations
   integer, parameter :: nsims = 10000
   integer, parameter :: ndrop = 500
   integer, parameter :: ncb = 2                         ! number of choice variables
   real (dp), parameter :: cmin = 0.001_dp

   ! Transition

   integer, parameter :: nper = 60
   real (dp), parameter :: pi = 3.141592654_dp
   real (dp), parameter :: eps(2)  = (/ 1.0e-8, 1.0e-6 /)! Convergence criteria
   real (dp), parameter :: smin(2) = (/   14.0,   25.0 /)  ! Lower bound on continuous grid 
   real (dp), parameter :: smax(2) = (/ 1900.0, 5200.0 /)  ! Upper bound on the continuous grid
   real (dp), parameter :: suv(2)  = (/   2.0,     2.0 /)  ! Degree of nonuniformity in spacing grid points
                                                           ! suv > 1 gives closer spacing near smin
   integer, parameter :: z0 = 1
   integer, parameter :: st0 = 1
   real (dp), parameter :: x0 = 0.5_dp*(smax(1)+smin(1))
   real (dp), parameter :: k0 = 0.5_dp*(smax(2)+smin(2))
 
   !-----------------------------------------------------!
   ! Global vars                                         !
   !-----------------------------------------------------!
   
   ! Workers

   integer :: nsmprdw(nsw),nsmsumw(nsw),nspw,nsdsumw
   integer, allocatable, dimension(:,:) :: jobw_mpi
   real (dp) :: ct,dt1,at1,lbd,rl,rs
   real (dp), allocatable, dimension(:) :: smw
   real (dp), allocatable, dimension(:) :: wbar
   real (dp), allocatable, dimension(:,:) :: worker_vars
   real (dp), allocatable, dimension(:,:) :: wtm
   real (dp), allocatable, dimension(:,:) :: hwm
   real (dp), allocatable, dimension(:,:) :: hwpm
   real (dp), allocatable, dimension(:,:) :: hdwt
   real (dp), allocatable, dimension(:,:) :: hawt
   real (dp), allocatable, dimension(:,:,:) :: wctp
   real (dp), allocatable, dimension(:,:,:) :: wdtp
   real (dp), allocatable, dimension(:,:,:) :: watp

   ! Entrepreneurs

   integer :: nsmprde(nse),nsmsume(nse),nspe,nsdsume
   integer, allocatable, dimension(:,:) :: jobe_mpi
   integer, allocatable, dimension(:) :: zesim
   real (dp) :: kt1,bt1,lt1
   real (dp) :: se(nse),se1(nse)
   real (dp) :: alphe1,delte1,ne1,cte1,xi,mui  
   real (dp), allocatable, dimension(:) :: sme
   real (dp), allocatable, dimension(:) :: ebar
   real (dp), allocatable, dimension(:,:) :: entre_vars
   real (dp), allocatable, dimension(:,:) :: etm
   real (dp), allocatable, dimension(:,:) :: hem
   real (dp), allocatable, dimension(:,:) :: hepm
   real (dp), allocatable, dimension(:,:) :: hket
   real (dp), allocatable, dimension(:,:) :: hset
   real (dp), allocatable, dimension(:,:) :: hbet
   real (dp), allocatable, dimension(:,:,:) :: ectp
   real (dp), allocatable, dimension(:,:,:) :: ektp
   real (dp), allocatable, dimension(:,:,:) :: ebtp
   real (dp), allocatable, dimension(:,:,:) :: eatp

   
   ! Bankers

   integer :: nsmprd(ns),nsmsum(ns),nsp,nsdsum
   integer :: k(ns)
   integer :: is,it
   integer :: lcr_on
   integer :: ires
   real (dp) :: s(ns),s1(ns)
   real (dp) :: rk,alph1,rb,grb,mgr,delta1,rt,wt,rd,rg,grg 
   real (dp) :: nu1,nu_b,xi0,nb1,ctb1
   real (dp) :: kt,bt,kp,xt,lp,adjc,bp,np,nt,alpha1
   real (dp) :: time_begin,time_end                      ! For the timing of the program
   real (dp), allocatable, dimension(:) :: sm
   real (dp), allocatable, dimension(:) :: kpgrd,igrid 
   real (dp), allocatable, dimension(:,:,:) :: vf,value
   real (dp), allocatable, dimension(:,:,:) :: g_k,g_s,g_l,g_c,g_m
   real (dp), allocatable, dimension(:) :: y
   real (dp), allocatable, dimension(:) :: st,dst
   real (dp), allocatable, dimension(:) :: exp_st
   real (dp), allocatable, dimension(:,:) :: banker_vars
   real (dp), allocatable, dimension(:,:) :: tm
   real (dp), allocatable, dimension(:,:) :: tm_st
   real (dp), allocatable, dimension(:,:) :: smins,smaxs
   real (dp), allocatable, dimension(:,:) :: sms
   integer, allocatable, dimension(:) :: zsim
   integer, allocatable, dimension(:) :: stsim
   integer, allocatable, dimension(:,:) :: jobs_mpi
   real (dp), allocatable, dimension(:,:,:,:) :: bctp
   real (dp), allocatable, dimension(:,:,:,:) :: bltp
   real (dp), allocatable, dimension(:,:,:,:) :: bdtp
   real (dp), allocatable, dimension(:,:,:,:) :: bstp

  ! Transitional dynamics

   integer :: nit
   real (dp) :: rl0,rl1,rs0,rs1,kl0,kl1,Sbar
   real (dp), dimension(nper) :: shockb,shocke,shockw
   real (dp), allocatable, dimension(:) :: rlvec,rsvec,klvec     ! vectors with the evolution of prices
   real (dp), allocatable, dimension(:) :: rlvec0,rsvec0,klvec0
   real (dp), allocatable, dimension(:) :: bloan,bsecur,blabor,sdepo,sdeponew
   real (dp), allocatable, dimension(:) :: wasset,dloan,easset
   real (dp), allocatable, dimension(:) :: dlabor
   real (dp), allocatable, dimension(:) :: loan_mkt
   real (dp), allocatable, dimension(:) :: whole_mkt
   real (dp), allocatable, dimension(:) :: klnew
   real (dp), allocatable, dimension(:,:) :: reg_req
 
   !-----------------------------------------------------!
   ! MPI variables                                       !
   !-----------------------------------------------------!

   integer :: ierr                                       ! return error message from mpi subroutines
   integer :: id                                         ! identification number of eqch processor
   integer :: nproc                                      ! total number of processors
   integer, parameter :: idmom=0                         ! node which works as frontend

end module mod_common_var




